@@ -36,9 +36,8 @@ import (
 )
 
 func init() {
-	// Ignore the value of CUE_EXPERIMENT for the purposes
-	// of these tests, which we want to test both with the experiment
-	// enabled and disabled.
+	// Ignore the value of CUE_EXPERIMENT for the purposes of these tests,
+	// as we want them to start off with the default experiment values.
 	os.Setenv("CUE_EXPERIMENT", "")
 
 	// Once we've called cueexperiment.Init, cueexperiment.Vars
@@ -132,7 +131,7 @@ module: mod.test/test@v0
 root:   $CWD/testdata/testmod
 dir:    ""
 display:""`}, {
-		name: "NoPackageName",
+		name: "NoMatchingPackageName",
 		cfg:  dirCfg,
 		args: []string{"./anon"},
 		want: `err:    build constraints exclude all CUE files in ./anon:
@@ -185,7 +184,7 @@ imports:
 		args: []string{"mod.test/test/hello:nonexist"},
 		want: `err:    cannot find package "mod.test/test/hello": no files in package directory with package name "nonexist"
 path:   mod.test/test/hello:nonexist
-module: mod.test/test@v0
+module: ""
 root:   $CWD/testdata/testmod
 dir:    ""
 display:mod.test/test/hello:nonexist`,
@@ -229,7 +228,7 @@ files:
 		want: `err:    cannot determine package name for "foo.com/bad-identifier"; set it explicitly with ':'
 cannot find package "foo.com/bad-identifier": cannot find module providing package foo.com/bad-identifier
 path:   foo.com/bad-identifier
-module: mod.test/test@v0
+module: ""
 root:   $CWD/testdata/testmod
 dir:    ""
 display:foo.com/bad-identifier`,
@@ -239,7 +238,7 @@ display:foo.com/bad-identifier`,
 		args: []string{"nonexisting"},
 		want: `err:    standard library import path "nonexisting" cannot be imported as a CUE package
 path:   nonexisting
-module: mod.test/test@v0
+module: ""
 root:   $CWD/testdata/testmod
 dir:    ""
 display:nonexisting`,
@@ -249,7 +248,7 @@ display:nonexisting`,
 		args: []string{"strconv"},
 		want: `err:    standard library import path "strconv" cannot be imported as a CUE package
 path:   strconv
-module: mod.test/test@v0
+module: ""
 root:   $CWD/testdata/testmod
 dir:    ""
 display:strconv`,
@@ -277,6 +276,20 @@ files:
 imports:
     mod.test/catch: $CWD/testdata/testmod/cue.mod/pkg/mod.test/catch/catch.cue
     mod.test/helper:helper1: $CWD/testdata/testmod/cue.mod/pkg/mod.test/helper/helper1.cue`}, {
+		name: "PackageWithImportsWithSkipImportsConfig",
+		cfg: &Config{
+			Dir:         testdataDir,
+			Tools:       true,
+			SkipImports: true,
+		},
+		args: []string{"./imports"},
+		want: `path:   mod.test/test/imports@v0
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/imports
+display:./imports
+files:
+    $CWD/testdata/testmod/imports/imports.cue`}, {
 		name: "OnlyToolFiles",
 		cfg:  dirCfg,
 		args: []string{"./toolonly"},
@@ -293,7 +306,6 @@ files:
 		},
 		args: []string{"./toolonly"},
 		want: `err:    build constraints exclude all CUE files in ./toolonly:
-    anon.cue: no package name
     test.cue: package is test, want foo
     toolonly/foo_tool.cue: _tool.cue files excluded in non-cmd mode
 path:   mod.test/test/toolonly@v0:foo
@@ -470,7 +482,176 @@ display:.
 files:
     $CWD/testdata/testmod/multi3/other.cue
 imports:
-    mod.test/test/sub: $CWD/testdata/testmod/sub/sub.cue`}}
+    mod.test/test/sub: $CWD/testdata/testmod/sub/sub.cue`}, {
+		// Test that we can explicitly ask for non-package
+		// CUE files by setting Config.Package to "_".
+		name: "ExplicitPackageWithUnqualifiedImportPath#4",
+		cfg: &Config{
+			Dir:     filepath.Join(testdataDir, "multi4"),
+			Package: "_",
+		},
+		args: []string{"."},
+		want: `path:   mod.test/test/multi4@v0:_
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/multi4
+display:.
+files:
+    $CWD/testdata/testmod/multi4/nopackage1.cue
+    $CWD/testdata/testmod/multi4/nopackage2.cue`}, {
+		// Test what happens when there's a single CUE file
+		// with an explicit `package _` directive.
+		name: "ExplicitPackageWithUnqualifiedImportPath#5",
+		cfg: &Config{
+			Dir:     filepath.Join(testdataDir, "multi5"),
+			Package: "_",
+		},
+		args: []string{"."},
+		want: `path:   mod.test/test/multi5@v0:_
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/multi5
+display:.
+files:
+    $CWD/testdata/testmod/multi5/nopackage.cue`}, {
+		// Check that imports are only considered from files
+		// that match the build paths.
+		name: "BuildTagsWithImports#1",
+		cfg: &Config{
+			Dir:  filepath.Join(testdataDir, "tagswithimports"),
+			Tags: []string{"prod"},
+		},
+		args: []string{"."},
+		want: `path:   mod.test/test/tagswithimports@v0
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/tagswithimports
+display:.
+files:
+    $CWD/testdata/testmod/tagswithimports/prod.cue
+imports:
+    mod.test/test/hello:test: $CWD/testdata/testmod/test.cue $CWD/testdata/testmod/hello/test.cue
+    mod.test/test/sub: $CWD/testdata/testmod/sub/sub.cue`}, {
+		// Check that imports are only considered from files
+		// that match the build paths. When we don't have the prod
+		// tag, the bad import path mentioned in testdata/testmod/tagswithimports/nonprod.cue
+		// surfaces in the errors.
+		name: "BuildTagsWithImports#2",
+		cfg: &Config{
+			Dir: filepath.Join(testdataDir, "tagswithimports"),
+		},
+		args: []string{"."},
+		want: `err:    mod.test/test/tagswithimports@v0: import failed: cannot find package "bad-import.example/foo": cannot find module providing package bad-import.example/foo:
+    $CWD/testdata/testmod/tagswithimports/nonprod.cue:5:8
+path:   mod.test/test/tagswithimports@v0
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/tagswithimports
+display:.
+files:
+    $CWD/testdata/testmod/tagswithimports/nonprod.cue`}, {
+		name: "ModuleFileNonDirectory",
+		cfg: &Config{
+			Dir: testdata("testmod_legacymodfile"),
+		},
+		args: []string{"."},
+		want: `err:    cue.mod files are no longer supported; use cue.mod/module.cue
+path:   ""
+module: ""
+root:   ""
+dir:    ""
+display:""`}, {
+		// This test checks that files in parent directories
+		// do not result in irrelevant instances appearing
+		// in the result of Instances.
+		name: "Issue3306",
+		cfg: &Config{
+			Dir:         testdataDir,
+			Package:     "*",
+			SkipImports: true,
+		},
+		args: []string{"./issue3306/..."},
+		want: `path:   mod.test/test/issue3306@v0:x
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/issue3306
+display:./issue3306
+files:
+    $CWD/testdata/testmod/issue3306/x.cue
+
+path:   mod.test/test/issue3306/a@v0:a
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/issue3306/a
+display:./issue3306/a
+files:
+    $CWD/testdata/testmod/issue3306/a/a.cue
+
+path:   mod.test/test/issue3306/a@v0:b
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/issue3306/a
+display:./issue3306/a
+files:
+    $CWD/testdata/testmod/issue3306/a/b.cue
+
+path:   mod.test/test/issue3306/a@v0:x
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/issue3306/a
+display:./issue3306/a
+files:
+    $CWD/testdata/testmod/issue3306/x.cue
+    $CWD/testdata/testmod/issue3306/a/x.cue
+
+path:   mod.test/test/issue3306/x@v0:x
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/issue3306/x
+display:./issue3306/x
+files:
+    $CWD/testdata/testmod/issue3306/x.cue
+    $CWD/testdata/testmod/issue3306/x/x.cue`}, {
+		// This test checks that when we use Package: "*",
+		// we can still use imported packages.
+		name: "AllPackagesWithImports",
+		cfg: &Config{
+			Dir:     testdataDir,
+			Package: "*",
+		},
+		args: []string{"."},
+		want: `path:   mod.test/test@v0:_
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod
+display:.
+files:
+    $CWD/testdata/testmod/anon.cue
+
+path:   mod.test/test@v0:test
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod
+display:.
+files:
+    $CWD/testdata/testmod/test.cue
+imports:
+    mod.test/test/sub: $CWD/testdata/testmod/sub/sub.cue`}, {
+		// This tests that we can load a CUE package by pointing Dir to it
+		// even when the package's directory name ends with ".cue".
+		name: "DirWithCUEFileExtension",
+		cfg: &Config{
+			Dir: filepath.Join(testdataDir, "testdir.cue"),
+		},
+		args: []string{"."},
+		want: `path:   mod.test/test/testdir.cue@v0:testdir
+module: mod.test/test@v0
+root:   $CWD/testdata/testmod
+dir:    $CWD/testdata/testmod/testdir.cue
+display:.
+files:
+    $CWD/testdata/testmod/testdir.cue/test.cue`,
+	}}
 	tdtest.Run(t, testCases, func(t *tdtest.T, tc *loadTest) {
 		pkgs := Instances(tc.args, tc.cfg)
 
@@ -589,7 +770,7 @@ package %s
 
 x: 1
 `, string(c))
-		err := os.WriteFile(filepath.Join(testDir, string(c)+".cue"), []byte(contents), 0o644)
+		err := os.WriteFile(filepath.Join(testDir, string(c)+".cue"), []byte(contents), 0o666)
 		qt.Assert(t, qt.IsNil(err))
 	}
 
