@@ -115,7 +115,7 @@ func (o *hiddenStructValue) at(i int) *adt.Vertex {
 	return o.arcs[i]
 }
 
-// Lookup reports the field for the given key. The returned Value is invalid
+// Lookup reports the field for the given key. The returned [Value] is invalid
 // if it does not exist.
 func (o *hiddenStructValue) Lookup(key string) Value {
 	f := o.v.idx.StrLabel(key)
@@ -127,7 +127,7 @@ func (o *hiddenStructValue) Lookup(key string) Value {
 		}
 	}
 	if i == len {
-		x := mkErr(o.v.idx, o.obj, 0, "field not found: %v", key)
+		x := mkErr(o.obj, 0, "field not found: %v", key)
 		x.NotExists = true
 		// TODO: more specifically we should test whether the values that
 		// are addressable from the root of the configuration can support the
@@ -143,22 +143,22 @@ func (o *hiddenStructValue) Lookup(key string) Value {
 
 // MarshalJSON returns a valid JSON encoding or reports an error if any of the
 // fields is invalid.
-func (o *structValue) marshalJSON() (b []byte, err errors.Error) {
+func (o *structValue) appendJSON(b []byte) ([]byte, error) {
 	b = append(b, '{')
 	n := o.Len()
 	for i := range n {
 		k, v := o.At(i)
+		// Do not use json.Marshal as it escapes HTML.
 		s, err := internaljson.Marshal(k)
 		if err != nil {
-			return nil, unwrapJSONError(err)
+			return nil, err
 		}
 		b = append(b, s...)
 		b = append(b, ':')
-		bb, err := internaljson.Marshal(v)
+		b, err = v.appendJSON(o.ctx, b)
 		if err != nil {
-			return nil, unwrapJSONError(err)
+			return nil, err
 		}
-		b = append(b, bb...)
 		if i < n-1 {
 			b = append(b, ',')
 		}
@@ -180,7 +180,7 @@ func toMarshalErr(v Value, b *adt.Bottom) error {
 
 func marshalErrf(v Value, src adt.Node, code adt.ErrorCode, msg string, args ...interface{}) error {
 	arguments := append([]interface{}{code, msg}, args...)
-	b := mkErr(v.idx, src, arguments...)
+	b := mkErr(src, arguments...)
 	return toMarshalErr(v, b)
 }
 
@@ -259,7 +259,7 @@ func (i *Iterator) Selector() Selector {
 // Label reports the label of the value if i iterates over struct fields and ""
 // otherwise.
 //
-// Slated to be deprecated: use [Iterator.Selector] and [Selector.String].
+// Deprecated: use [Iterator.Selector] and [Selector.String].
 // Note that this will give more accurate string representations.
 func (i *hiddenIterator) Label() string {
 	if i.f == 0 {
@@ -268,13 +268,6 @@ func (i *hiddenIterator) Label() string {
 	return i.idx.LabelStr(i.f)
 }
 
-// IsHidden reports if a field is hidden from the data model.
-//
-// Deprecated: use i.Selector().PkgPath() != ""
-func (i *hiddenIterator) IsHidden() bool {
-	return i.f.IsHidden()
-}
-
 // IsOptional reports if a field is optional.
 func (i *Iterator) IsOptional() bool {
 	return i.arcType == adt.ArcOptional
@@ -285,24 +278,17 @@ func (i *Iterator) FieldType() SelectorType {
 	return featureToSelType(i.f, i.arcType)
 }
 
-// IsDefinition reports if a field is a definition.
-//
-// Deprecated: use i.Selector().IsDefinition()
-func (i *hiddenIterator) IsDefinition() bool {
-	return i.f.IsDef()
-}
-
 // marshalJSON iterates over the list and generates JSON output. HasNext
 // will return false after this operation.
-func marshalList(l *Iterator) (b []byte, err errors.Error) {
+func listAppendJSON(b []byte, l *Iterator) ([]byte, error) {
 	b = append(b, '[')
 	if l.Next() {
 		for i := 0; ; i++ {
-			x, err := internaljson.Marshal(l.Value())
+			var err error
+			b, err = l.Value().appendJSON(l.ctx, b)
 			if err != nil {
-				return nil, unwrapJSONError(err)
+				return nil, err
 			}
-			b = append(b, x...)
 			if !l.Next() {
 				break
 			}
@@ -733,133 +719,6 @@ func (v Value) Default() (Value, bool) {
 		return v, false
 	}
 	return makeValue(v.idx, d, v.parent_), true
-
-	// d, ok := v.v.Value.(*adt.Disjunction)
-	// if !ok {
-	// 	return v, false
-	// }
-
-	// var w *adt.Vertex
-
-	// switch d.NumDefaults {
-	// case 0:
-	// 	return v, false
-
-	// case 1:
-	// 	w = d.Values[0]
-
-	// default:
-	// 	x := *v.v
-	// 	x.Value = &adt.Disjunction{
-	// 		Src:         d.Src,
-	// 		Values:      d.Values[:d.NumDefaults],
-	// 		NumDefaults: 0,
-	// 	}
-	// 	w = &x
-	// }
-
-	// w.Conjuncts = nil
-	// for _, c := range v.v.Conjuncts {
-	// 	// TODO: preserve field information.
-	// 	expr, _ := stripNonDefaults(c.Expr())
-	// 	w.AddConjunct(adt.MakeConjunct(c.Env, expr))
-	// }
-
-	// return makeValue(v.idx, w), true
-
-	// if !stripped {
-	// 	return v, false
-	// }
-
-	// n := *v.v
-	// n.Conjuncts = conjuncts
-	// return Value{v.idx, &n}, true
-
-	// isDefault := false
-	// for _, c := range v.v.Conjuncts {
-	// 	if hasDisjunction(c.Expr()) {
-	// 		isDefault = true
-	// 		break
-	// 	}
-	// }
-
-	// if !isDefault {
-	// 	return v, false
-	// }
-
-	// TODO: record expanded disjunctions in output.
-	// - Rename Disjunction to DisjunctionExpr
-	// - Introduce Disjuncts with Values.
-	// - In Expr introduce Star
-	// - Don't pick default by default?
-
-	// Evaluate the value.
-	// 	x := eval.FinalizeValue(v.idx.Runtime, v.v)
-	// 	if b, _ := x.Value.(*adt.Bottom); b != nil { // && b.IsIncomplete() {
-	// 		return v, false
-	// 	}
-	// 	// Finalize and return here.
-	// 	return Value{v.idx, x}, isDefault
-}
-
-// TODO: this should go: record preexpanded disjunctions in Vertex.
-func hasDisjunction(expr adt.Expr) bool {
-	switch x := expr.(type) {
-	case *adt.DisjunctionExpr:
-		return true
-	case *adt.Conjunction:
-		for _, v := range x.Values {
-			if hasDisjunction(v) {
-				return true
-			}
-		}
-	case *adt.BinaryExpr:
-		switch x.Op {
-		case adt.OrOp:
-			return true
-		case adt.AndOp:
-			return hasDisjunction(x.X) || hasDisjunction(x.Y)
-		}
-	}
-	return false
-}
-
-// TODO: this should go: record preexpanded disjunctions in Vertex.
-func stripNonDefaults(expr adt.Expr) (r adt.Expr, stripped bool) {
-	switch x := expr.(type) {
-	case *adt.DisjunctionExpr:
-		if !x.HasDefaults {
-			return x, false
-		}
-		d := *x
-		d.Values = []adt.Disjunct{}
-		for _, v := range x.Values {
-			if v.Default {
-				d.Values = append(d.Values, v)
-			}
-		}
-		if len(d.Values) == 1 {
-			return d.Values[0].Val, true
-		}
-		return &d, true
-
-	case *adt.BinaryExpr:
-		if x.Op != adt.AndOp {
-			return x, false
-		}
-		a, sa := stripNonDefaults(x.X)
-		b, sb := stripNonDefaults(x.Y)
-		if sa || sb {
-			bin := *x
-			bin.X = a
-			bin.Y = b
-			return &bin, true
-		}
-		return x, false
-
-	default:
-		return x, false
-	}
 }
 
 // Label reports he label used to obtain this value from the enclosing struct.
@@ -880,12 +739,9 @@ func (v Value) Kind() Kind {
 	if v.v == nil {
 		return BottomKind
 	}
-	c := v.v.BaseValue
-	if !v.v.IsConcrete() {
-		return BottomKind
-	}
-	// TODO: perhaps we should not consider open lists as "incomplete".
-	if v.IncompleteKind() == adt.ListKind && !v.v.IsClosedList() {
+	w := v.v.DerefValue()
+	c := w.BaseValue
+	if !w.IsConcrete() {
 		return BottomKind
 	}
 	return c.Kind()
@@ -901,19 +757,19 @@ func (v Value) IncompleteKind() Kind {
 
 // MarshalJSON marshalls this value into valid JSON.
 func (v Value) MarshalJSON() (b []byte, err error) {
-	b, err = v.marshalJSON()
+	ctx := newContext(v.idx)
+	b, err = v.appendJSON(ctx, nil)
 	if err != nil {
 		return nil, unwrapJSONError(err)
 	}
 	return b, nil
 }
 
-func (v Value) marshalJSON() (b []byte, err error) {
+func (v Value) appendJSON(ctx *adt.OpContext, b []byte) ([]byte, error) {
 	v, _ = v.Default()
 	if v.v == nil {
-		return internaljson.Marshal(nil)
+		return append(b, "null"...), nil
 	}
-	ctx := newContext(v.idx)
 	x := v.eval(ctx)
 
 	if _, ok := x.(adt.Resolver); ok {
@@ -926,26 +782,32 @@ func (v Value) marshalJSON() (b []byte, err error) {
 	// TODO: implement marshalles in value.
 	switch k := x.Kind(); k {
 	case adt.NullKind:
-		return internaljson.Marshal(nil)
+		return append(b, "null"...), nil
 	case adt.BoolKind:
-		return internaljson.Marshal(x.(*adt.Bool).B)
+		b2, err := json.Marshal(x.(*adt.Bool).B)
+		return append(b, b2...), err
 	case adt.IntKind, adt.FloatKind, adt.NumberKind:
-		b, err := x.(*adt.Num).X.MarshalText()
-		b = bytes.TrimLeft(b, "+")
-		return b, err
+		// TODO(mvdan): MarshalText does not guarantee valid JSON,
+		// but apd.Decimal does not expose a MarshalJSON method either.
+		b2, err := x.(*adt.Num).X.MarshalText()
+		b2 = bytes.TrimLeft(b2, "+")
+		return append(b, b2...), err
 	case adt.StringKind:
-		return internaljson.Marshal(x.(*adt.String).Str)
+		// Do not use json.Marshal as it escapes HTML.
+		b2, err := internaljson.Marshal(x.(*adt.String).Str)
+		return append(b, b2...), err
 	case adt.BytesKind:
-		return internaljson.Marshal(x.(*adt.Bytes).B)
+		b2, err := json.Marshal(x.(*adt.Bytes).B)
+		return append(b, b2...), err
 	case adt.ListKind:
-		i, _ := v.List()
-		return marshalList(&i)
+		i := v.mustList(ctx)
+		return listAppendJSON(b, &i)
 	case adt.StructKind:
 		obj, err := v.structValData(ctx)
 		if err != nil {
 			return nil, toMarshalErr(v, err)
 		}
-		return obj.marshalJSON()
+		return obj.appendJSON(b)
 	case adt.BottomKind:
 		return nil, toMarshalErr(v, x.(*adt.Bottom))
 	default:
@@ -962,8 +824,7 @@ func (v Value) Syntax(opts ...Option) ast.Node {
 	if v.v == nil {
 		return nil
 	}
-	var o options = getOptions(opts)
-	// var inst *Instance
+	o := getOptions(opts)
 
 	p := export.Profile{
 		Simplify:        !o.raw,
@@ -1049,26 +910,6 @@ func (v Value) Doc() []*ast.CommentGroup {
 	return export.ExtractDoc(v.v)
 }
 
-// Split returns a list of values from which v originated such that
-// the unification of all these values equals v and for all returned values.
-// It will also split unchecked unifications (embeddings), so unifying the
-// split values may fail if actually unified.
-// Source returns a non-nil value.
-//
-// Deprecated: use [Value.Expr].
-func (v hiddenValue) Split() []Value {
-	if v.v == nil {
-		return nil
-	}
-	a := []Value{}
-	v.v.VisitLeafConjuncts(func(x adt.Conjunct) bool {
-		env, expr := x.EnvExpr()
-		a = append(a, remakeValue(v, env, expr))
-		return true
-	})
-	return a
-}
-
 // Source returns the original node for this value. The return value may not
 // be an [ast.Expr]. For instance, a struct kind may be represented by a
 // struct literal, a field comprehension, or a file. It returns nil for
@@ -1093,8 +934,7 @@ func (v Value) Source() ast.Node {
 // If v exactly represents a package, BuildInstance returns
 // the build instance corresponding to the value; otherwise it returns nil.
 //
-// The value returned by Value.ReferencePath will commonly represent
-// a package.
+// The value returned by [Value.ReferencePath] will commonly represent a package.
 func (v Value) BuildInstance() *build.Instance {
 	if v.idx == nil {
 		return nil
@@ -1112,7 +952,7 @@ func (v Value) Err() error {
 
 // Pos returns position information.
 //
-// Use v.Expr to get positions for all conjuncts and disjuncts.
+// Use [Value.Expr] to get positions for all conjuncts and disjuncts.
 func (v Value) Pos() token.Pos {
 	if v.v == nil {
 		return token.NoPos
@@ -1183,13 +1023,11 @@ func (v Value) IsConcrete() bool {
 	if v.v == nil {
 		return false // any is neither concrete, not a list or struct.
 	}
-	if b := v.v.Bottom(); b != nil {
+	w := v.v.DerefValue()
+	if b := w.Bottom(); b != nil {
 		return !b.IsIncomplete()
 	}
-	if !adt.IsConcrete(v.v) {
-		return false
-	}
-	if v.IncompleteKind() == adt.ListKind && !v.v.IsClosedList() {
+	if !adt.IsConcrete(w) {
 		return false
 	}
 	return true
@@ -1251,11 +1089,11 @@ func (v Value) checkKind(ctx *adt.OpContext, want adt.Kind) *adt.Bottom {
 	k := x.Kind()
 	if want != adt.BottomKind {
 		if k&want == adt.BottomKind {
-			return mkErr(v.idx, x, "cannot use value %v (type %s) as %s",
+			return mkErr(x, "cannot use value %v (type %s) as %s",
 				ctx.Str(x), k, want)
 		}
 		if !adt.IsConcrete(x) {
-			return mkErr(v.idx, x, adt.IncompleteError, "non-concrete value %v", k)
+			return mkErr(x, adt.IncompleteError, "non-concrete value %v", k)
 		}
 	}
 	return nil
@@ -1298,13 +1136,13 @@ func (v Value) Len() Value {
 		}
 	}
 	const msg = "len not supported for type %v"
-	return remakeValue(v, nil, mkErr(v.idx, v.v, msg, v.Kind()))
+	return remakeValue(v, nil, mkErr(v.v, msg, v.Kind()))
 
 }
 
 // Elem returns the value of undefined element types of lists and structs.
 //
-// Deprecated: use [Value.LookupPath] in combination with "AnyString" or "AnyIndex".
+// Deprecated: use [Value.LookupPath] in combination with [AnyString] or [AnyIndex].
 func (v hiddenValue) Elem() (Value, bool) {
 	sel := AnyString
 	if v.v.IsList() {
@@ -1322,13 +1160,19 @@ func (v Value) List() (Iterator, error) {
 	if err := v.checkKind(ctx, adt.ListKind); err != nil {
 		return Iterator{idx: v.idx, ctx: ctx}, v.toErr(err)
 	}
+	return v.mustList(ctx), nil
+}
+
+// mustList is like [Value.List], but reusing ctx and leaving it to the caller
+// to apply defaults and check the kind.
+func (v Value) mustList(ctx *adt.OpContext) Iterator {
 	arcs := []*adt.Vertex{}
 	for _, a := range v.v.Elems() {
 		if a.Label.IsInt() {
 			arcs = append(arcs, a)
 		}
 	}
-	return Iterator{idx: v.idx, ctx: ctx, val: v, arcs: arcs}, nil
+	return Iterator{idx: v.idx, ctx: ctx, val: v, arcs: arcs}
 }
 
 // Null reports an error if v is not null.
@@ -1419,8 +1263,7 @@ func (v Value) structValOpts(ctx *adt.OpContext, o options) (s structValue, err
 	// Allow scalar values if hidden or definition fields are requested.
 	case !o.omitHidden, !o.omitDefinitions:
 	default:
-		obj, err = v.getStruct()
-		if err != nil {
+		if err := v.checkKind(ctx, adt.StructKind); err != nil && !err.ChildError {
 			return structValue{}, err
 		}
 	}
@@ -1482,17 +1325,9 @@ func (v hiddenValue) Struct() (*Struct, error) {
 	return &Struct{obj}, nil
 }
 
-func (v Value) getStruct() (*adt.Vertex, *adt.Bottom) {
-	ctx := v.ctx()
-	if err := v.checkKind(ctx, adt.StructKind); err != nil {
-		if !err.ChildError {
-			return nil, err
-		}
-	}
-	return v.v, nil
-}
-
 // Struct represents a CUE struct value.
+//
+// Deprecated: only used by deprecated functions.
 type Struct struct {
 	structValue
 }
@@ -1545,7 +1380,7 @@ func (s *hiddenStruct) FieldByName(name string, isIdent bool) (FieldInfo, error)
 	return FieldInfo{}, errNotFound
 }
 
-// Fields creates an iterator over the Struct's fields.
+// Fields creates an iterator over the struct's fields.
 func (s *hiddenStruct) Fields(opts ...Option) *Iterator {
 	iter, _ := s.v.Fields(opts...)
 	return iter
@@ -1716,15 +1551,15 @@ func (v hiddenValue) Fill(x interface{}, path ...string) Value {
 // FillPath creates a new value by unifying v with the value of x at the given
 // path.
 //
-// If x is an cue/ast.Expr, it will be evaluated within the context of the
+// If x is an [ast.Expr], it will be evaluated within the context of the
 // given path: identifiers that are not resolved within the expression are
 // resolved as if they were defined at the path position.
 //
 // If x is a Value, it will be used as is. It panics if x is not created
-// from the same Runtime as v.
+// from the same [Context] as v.
 //
 // Otherwise, the given Go value will be converted to CUE using the same rules
-// as Context.Encode.
+// as [Context.Encode].
 //
 // Any reference in v referring to the value at the given path will resolve to x
 // in the newly created value. The resulting value is not validated.
@@ -1735,7 +1570,7 @@ func (v Value) FillPath(p Path, x interface{}) Value {
 	}
 	ctx := v.ctx()
 	if err := p.Err(); err != nil {
-		return newErrValue(v, mkErr(v.idx, nil, 0, "invalid path: %v", err))
+		return newErrValue(v, mkErr(nil, 0, "invalid path: %v", err))
 	}
 	var expr adt.Expr
 	switch x := x.(type) {
@@ -1858,22 +1693,6 @@ func (v Value) Subsume(w Value, opts ...Option) error {
 	return p.Value(ctx, v.v, w.v)
 }
 
-// Deprecated: use [Value.Subsume].
-//
-// Subsumes reports whether w is an instance of v.
-//
-// Without options, Subsumes checks whether v is a backwards compatible schema
-// of w.
-//
-// By default, Subsumes tests whether two values are compatible
-// Value v and w must be obtained from the same build.
-// TODO: remove this requirement.
-func (v hiddenValue) Subsumes(w Value) bool {
-	ctx := v.ctx()
-	p := subsume.Profile{Defaults: true}
-	return p.Check(ctx, v.v, w.v)
-}
-
 func allowed(ctx *adt.OpContext, parent, n *adt.Vertex) *adt.Bottom {
 	if !parent.IsClosedList() && !parent.IsClosedStruct() {
 		return nil
@@ -1892,9 +1711,9 @@ func allowed(ctx *adt.OpContext, parent, n *adt.Vertex) *adt.Bottom {
 
 func addConjuncts(dst, src *adt.Vertex) {
 	c := adt.MakeRootConjunct(nil, src)
-	if src.Closed {
+	if src.ClosedRecursive {
 		var root adt.CloseInfo
-		c.CloseInfo = root.SpawnRef(src, src.Closed, nil)
+		c.CloseInfo = root.SpawnRef(src, src.ClosedRecursive, nil)
 	}
 	dst.AddConjunct(c)
 }
@@ -1920,7 +1739,7 @@ func (v Value) Unify(w Value) Value {
 
 	n.Parent = v.v.Parent
 	n.Label = v.v.Label
-	n.Closed = v.v.Closed || w.v.Closed
+	n.ClosedRecursive = v.v.ClosedRecursive || w.v.ClosedRecursive
 
 	if err := n.Err(ctx); err != nil {
 		return makeValue(v.idx, n, v.parent_)
@@ -1935,7 +1754,7 @@ func (v Value) Unify(w Value) Value {
 	return makeValue(v.idx, n, v.parent_)
 }
 
-// UnifyAccept is as v.Unify(w), but will disregard the closedness rules for
+// UnifyAccept is like [Value.Unify](w), but will disregard the closedness rules for
 // v and w, and will, instead, only allow fields that are present in accept.
 //
 // UnifyAccept is used to piecemeal unify individual conjuncts obtained from
@@ -2013,7 +1832,7 @@ func (v hiddenValue) Reference() (inst *Instance, path []string) {
 }
 
 // ReferencePath returns the value and path referred to by this value such that
-// value.LookupPath(path) resolves to the same value, or no path if this value
+// [Value.LookupPath](path) resolves to the same value, or no path if this value
 // is not a reference.
 func (v Value) ReferencePath() (root Value, p Path) {
 	// TODO: don't include references to hidden fields.
@@ -2260,7 +2079,7 @@ func (o *options) updateOptions(opts []Option) {
 }
 
 // Validate reports any errors, recursively. The returned error may represent
-// more than one error, retrievable with errors.Errors, if more than one
+// more than one error, retrievable with [errors.Errors], if more than one
 // exists.
 //
 // Note that by default not all errors are reported, unless options like
