@@ -93,7 +93,10 @@ func (n *nodeContext) insertConstraint(pattern Value, c Conjunct) bool {
 	}
 
 	if constraint == nil {
-		constraint = &Vertex{}
+		constraint = &Vertex{
+			// See "Self-referencing patterns" in cycle.go
+			IsPatternConstraint: true,
+		}
 		pcs.Pairs = append(pcs.Pairs, PatternConstraint{
 			Pattern:    pattern,
 			Constraint: constraint,
@@ -109,6 +112,8 @@ func (n *nodeContext) insertConstraint(pattern Value, c Conjunct) bool {
 
 // matchPattern reports whether f matches pattern. The result reflects
 // whether unification of pattern with f converted to a CUE value succeeds.
+// The caller should check separately whether f matches any other arcs
+// that are not covered by pattern.
 func matchPattern(ctx *OpContext, pattern Value, f Feature) bool {
 	if pattern == nil || !f.IsRegular() {
 		return false
@@ -134,6 +139,9 @@ func matchPattern(ctx *OpContext, pattern Value, f Feature) bool {
 // This is an optimization an intended to be faster than regular CUE evaluation
 // for the majority of cases where pattern constraints are used.
 func matchPatternValue(ctx *OpContext, pattern Value, f Feature, label Value) (result bool) {
+	if v, ok := pattern.(*Vertex); ok {
+		v.unify(ctx, scalarKnown, finalize)
+	}
 	pattern = Unwrap(pattern)
 	label = Unwrap(label)
 
@@ -155,11 +163,13 @@ func matchPatternValue(ctx *OpContext, pattern Value, f Feature, label Value) (r
 		// TODO: hoist and reuse with the identical code in optional.go.
 		if x == cycle {
 			err := ctx.NewPosf(pos(pattern), "cyclic pattern constraint")
-			for _, c := range ctx.vertex.Conjuncts {
+			ctx.vertex.VisitLeafConjuncts(func(c Conjunct) bool {
 				addPositions(err, c)
-			}
+				return true
+			})
 			ctx.AddBottom(&Bottom{
-				Err: err,
+				Err:  err,
+				Node: ctx.vertex,
 			})
 		}
 		if ctx.errs == nil {
